SystemOrganization addCategory: #'Xtreams-Transforms'!

XTWriteStream subclass: #XTCollectWriteStream
	instanceVariableNames: 'block cache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTCollectWriteStream commentStamp: '<historical>' prior: 0!
Converts written elements using the provided conversion block. Optimizes bulk writes by batching the conversion results, allowing the destination stream to perform bulk writes as well.

Instance Variables
	block	<BlockClosure> collecting block (same style as collect: blocks)
	cache	<SequenceableCollection> caches results after transformation so that they can be bulk written too

!

!XTCollectWriteStream classSide methodsFor: 'instance creation' stamp: 'nice 10/10/2010 14:02'!
on: aWriteStream block: block
	^self new on: aWriteStream block: block! !

!XTCollectWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:53'!
++ anInteger
	^destination ++ anInteger! !

!XTCollectWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:53'!
-- anInteger
	^destination -- anInteger! !

!XTCollectWriteStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:25'!
block
	^block! !

!XTCollectWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:53'!
close
	super close.
	cache recycle.
	cache := nil! !

!XTCollectWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:56'!
contentsSpecies

	^cache class! !

!XTCollectWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:56'!
contentsSpecies: aClass

	cache ifNotNil: [ cache recycle ].
	cache := aClass newRecycled: (
				(cache notNil and: [ cache size > 0 ])
					ifTrue: [cache size]
					ifFalse: [DefaultBufferSize])! !

!XTCollectWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
insert: anInteger from: aSequenceableCollection at: startIndex
	| count amount written offset |
	count := 0.
	offset := startIndex.
	[count < anInteger] whileTrue:
		[amount := anInteger min: cache size.
		1 to: amount do: [:index | cache at: index put: (block value: (aSequenceableCollection at: offset+index))].
		written := [destination insert: anInteger from: cache. amount] on: Incomplete do: [ :ex | ex count ].
		count := count + written.
		written < amount ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise].
		offset := offset + amount ]! !

!XTCollectWriteStream methodsFor: 'testing' stamp: 'nice 10/10/2010 14:28'!
isPositionable
	^destination isPositionable! !

!XTCollectWriteStream methodsFor: 'seeking' stamp: 'nice 10/10/2010 14:28'!
length
	^destination length! !

!XTCollectWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:59'!
on: aDestination block: aBlock
	self on: aDestination.
	block := aBlock.
	self contentsSpecies: aDestination contentsSpecies! !

!XTCollectWriteStream methodsFor: 'seeking' stamp: 'nice 10/10/2010 14:28'!
position
	^destination position! !

!XTCollectWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:59'!
position: anInteger
	^destination position: anInteger! !

!XTCollectWriteStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:25'!
put: anObject
	destination put: (block value: anObject).
	^anObject! !

!XTCollectWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
write: anInteger from: aSequenceableCollection at: startIndex
	| count amount written offset |
	count := 0.
	offset := startIndex - 1.
	[count < anInteger] whileTrue:
		[amount := anInteger - count min: cache size.
		1 to: amount do: [:index | cache at: index put: (block value: (aSequenceableCollection at: offset + index))].
		written := [destination write: amount from: cache. amount] on: Incomplete do: [ :ex | ex count ].
		count := count + written.
		written < amount ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise].
		offset := offset + amount ].
	^anInteger! !

XTWriteStream subclass: #XTTransformWriteStream
	instanceVariableNames: 'buffer block closeBlock process incompleteCount readReady writeReady closeReady'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTTransformWriteStream commentStamp: '<historical>' prior: 0!
Transform write stream provides the most expressive form of transformation. The transformation is described by a binary block that is given two arguments, @input and @output. The @input is a virtual stream of elements written into the stream. The @output is the destination stream under the transform stream. The block can read arbitrary amount of elements from @input (including none) and write arbitrary amount of elements into @output (including none). The block will be invoked as many times as necessary to consume everything written into the stream, or until an Incomplete is raised by the destination.

The closing behavior of the stream can be customized through the closeBlock. The default closeBlock simply propagates #close to the destination as with any other transform stream.

In some cases it might be desirable to tweak the buffering strategy of the virtual input stream. In that case the buffer of the stream can be set manually to any kind of Buffer that matches the requirements of the transformation.

From the point of view of the API, the TransformWriteStream is very much like the TransformReadStream. Notably any valid transform block should work the same way on either read or write stream without modification. However to preserve the invariants and expressivity of the transform block, the implementation is vastly different. Primarily it is necessary to convert all the writes into a virtual stream of written elements that can be passed into the transform block as the input stream. Consequently the transformation itself needs to be suspended if there weren't enough elements written yet, to complete an iteration of the transform block. Therefore it needs to run in its own process. Any writes get redirected into an internal buffer and a background process repeatedly invokes the transform block to drain the contents of the buffer and produces output into the destination. Obviously buffer access has to be synchronized between any writing threads and the background process. The readReady/writeReady semaphores work in a lock-step mode, to interleave the background buffer draining with any writes.

Instance Variables
	buffer	<Buffer> holds the contents of the virtual input stream
	block	<BlockClosure> binary transformation block that reads elements from input (first argument) and writes elements into output (second argument)
	closeBlock	<BlockClosure> binary block invoked in response to the #close message, allows customizing the close behavior
	process	<Process> background process that runs the transformation block
	incompleteCount	<Integer> indicates that the transformation raised or received Incomplete and how many elements were actually consumed by the transformation block so that we can reraise Incomplete with correct count in the client thread
	readReady	<Semaphore> signals to the background process that elements were written into the buffer
	writeReady	<Semaphore> gates any writes into the stream, making sure background process is not draining the buffer at the same time
	closeReady	<Semaphore> signals back to the user thread that the background process finished draining the buffer and the stream is properly closed, so the #close call can return

!

!XTTransformWriteStream classSide methodsFor: 'instance creation' stamp: 'nice 10/10/2010 14:29'!
on: aWriteStream block: block
	^self new on: aWriteStream block: block! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:29'!
block
	^block! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:30'!
buffer
	^buffer! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:31'!
buffer: aBuffer
	buffer := aBuffer.
	self drainBuffer! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/31/2010 00:03'!
close
	buffer ifNil: [^self].
	"Write nothing to the buffer, but signal that there is data to read, causing the drain to throw Incomplete"
	readReady signal.

	"Wait for the drain to finish up completely. Rearm the closeReady incase somebody calls #close again."
	closeReady wait.
	closeReady signal.

	closeBlock cull: destination cull: self.

	buffer recycle.
	buffer := nil! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:30'!
closeBlock
	^closeBlock! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:32'!
closeBlock: anObject
	closeBlock := anObject! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:32'!
contentsSpecies
	^buffer contentsSpecies! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:32'!
contentsSpecies: aClass

	self buffer: (buffer class new: buffer cacheSize class: aClass)! !

!XTTransformWriteStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
drainBuffer
	| reading |
	process ifNotNil: [process terminate].
	incompleteCount == nil ifFalse: [^self].

	closeReady := Semaphore new.
	readReady := Semaphore new.
	writeReady := Semaphore new.
	reading := buffer reading transforming: [:in :out |
		| count |
		"Wait for the main process to have written to the buffer"
		readReady wait.

		"If nothing was written to the buffer, this indicates we're closing, we raise Incomplete"
		(count := buffer readSize) isZero ifTrue: [Incomplete zero raise].
		out write: count from: in.

		"Allow the main process to write to the buffer"
		writeReady signal].
	process := [
		[[block value: reading value: destination] repeat] on: Incomplete do: [].
		incompleteCount := reading buffer readPosition.

		"Signal that we're closed so that an attempt to write or close will immediately complete"
		writeReady signal.
		closeReady signal]
			newProcess.
	process resume! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
insert: anInteger from: aSequenceableCollection at: startIndex
	| count amount |
	anInteger isZero ifTrue: [^self].

	count := 0.
	[count < anInteger] whileTrue:
		[amount := anInteger - count.
		buffer hasFixedWriteSpace ifTrue: [amount := amount min: buffer cacheSize].
		buffer insert: amount from: aSequenceableCollection at: startIndex + count.

		"There is now data in the buffer for the drain to read"
		readReady signal.
		writeReady wait.
		incompleteCount == nil ifFalse: [(Incomplete count: count + incompleteCount) raise].
		count := count + amount ]! !

!XTTransformWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:33'!
on: aStreamable block: aBlock
	super on: aStreamable.
	block := aBlock.
	incompleteCount := nil.
	closeBlock := [ :destinationStream | destinationStream close].
	buffer := XTElasticBuffer new: DefaultBufferSize class: aStreamable contentsSpecies.
	self drainBuffer! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
put: anObject
	incompleteCount == nil ifFalse: [Incomplete zero raise].

	buffer put: anObject.

	"There is now data in the buffer for the drain to read"
	readReady signal.
	writeReady wait! !

!XTTransformWriteStream methodsFor: 'printing' stamp: 'nice 10/10/2010 14:34'!
streamingPrintOn: aStream
	super streamingPrintOn: aStream.
	aStream
		space;
		write: '#';
		write: block method homeMethod selector;
		write: ' buffered: ';
		print: buffer writeSize.
	buffer writeSize isZero ifTrue: [^self].
	aStream
		cr; tab;
		print: buffer contentsPast! !

!XTTransformWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
write: anInteger from: aSequenceableCollection at: startIndex
	| count amount |
	anInteger isZero ifTrue: [^0].

	count := 0.
	[count < anInteger] whileTrue:
		[amount := anInteger - count.
		buffer hasFixedWriteSpace ifTrue: [amount := amount min: buffer cacheSize].
		buffer write: amount from: aSequenceableCollection at: startIndex + count.
		
		"There is now data in the buffer for the drain to read"
		readReady signal.
		writeReady wait.
		incompleteCount == nil ifFalse: [(Incomplete count: count + incompleteCount) raise].
		count := count + amount ].
	^anInteger! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-private' stamp: 'nice 10/12/2010 00:44'!
basicNextPut: aCharacter
	"This has to be tricky: Squeak TextConverter will encode a Character into Character rather than anInteger..."
	self put: aCharacter asciiValue .
	^aCharacter! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:43'!
collecting: aBlock
	"Transform each written element using #collect: style block."
	"	aBlock	<BlockClosure>	a #collect: style block used to tranform each element
		^<CollectWriteSteam>
	""
		(Array new writing collecting: [ :e | e * e ]) write: (1 to: 5); conclusion
	""
		(String new writing collecting: [ :e | e asCharacter ]) write: (65 to: 90); conclusion
	"
	^XTCollectWriteStream on: self block: aBlock! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:43'!
doing: aBlock
	"Perform and action with each passing element using #do: style block."
	"	aBlock	<BlockClosure>	a #do: style block invoked with each element as it passes through the stream
		^<CollectWriteSteam>
	""
		(Array new writing doing: [ :e | Transcript space; print: e * e ]) write: (1 to: 10); conclusion
	"
	^self collecting: [:each | (aBlock value: each). each]! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/11/2010 22:05'!
encoding: anEncoding
	"Transform characters into bytes using @anEncoding such as #utf8 or #ascii, etc. Any encoding supported by StreamEncoder is allowed.
	The encoding steam also performs automatic conversion of CRs into the native line-end convention of the underlying platform,
	unless set into a different line-end convention mode"
	"	anEncoding	<Symbol> encoding identifier recognized by StreamEncoder class>>new:
		^<EncodedWriteStream>
	""
		(ByteArray new writing encoding: #ascii) write: 'abcdefghi'; conclusion
	""
		(ByteArray new writing encoding: #ascii) write: 'Hello\World' withCRs; conclusion
	""
		(ByteArray new writing encoding: #ascii) setLineEndCRLF; write: 'Hello\World' withCRs; conclusion
	"
	^XTEncodeWriteStream on: self encoding: anEncoding! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'MartinKobetic 1/11/2011 17:53'!
encodingBase64
	"Encodes bytes into characters of base-64 encoding.
	Emits final padding characters ($=) as required, when the stream is closed."
	"	^<TransformWriteStream>"
	"
		String new writing encodingBase64 write: (ByteArray withAll: (1 to: 20)); conclusion
	"
	| map cache |
	map := [ :i | 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/' at: i + 1 ].
	cache := ByteArray new: 3.
	^(self transforming: [ :in :out | | count block shift |
		count := [ in read: 3 into: cache at: 1. 3 ] on: Incomplete do: [ :incomplete | incomplete count].
		count isZero ifTrue: [ Incomplete zero raise ].
		block := (1 to: count) inject: 0 into: [ :total :byte | (total bitShift: 8) + (cache at: byte)].
		shift := count * -8.
		1 to: count + 1 do: [:i | out put: (map value: ((block bitShift: (shift + (i * 6))) bitAnd: 63))].
		count < 3 ifTrue: [
			3 - count timesRepeat: [ out put: $= ].
			(Incomplete count: count) raise]])
		buffer: (XTRingBuffer on: (ByteArray new: 3));
		yourself! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'MartinKobetic 1/17/2011 10:21'!
encodingHex
	"Encodes hex characters into bytes."
	"	^<TransformReadStream>"
	"
		ByteArray new writing encodingHex write: '010203fdfeff'; close; terminal
	"
	| c2i |
	c2i := [ :c | ('0123456789abcdef' indexOf: c asLowercase) - 1 ].
	^(self transforming: [ :in :out |
		out put: ((c2i value: in get) bitShift: 4) + (c2i value: in get) ])
		contentsSpecies: ByteString;
		yourself! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:44'!
injecting: initialObject into: aBlock
	"Accumulates a running value combined with each passing element using the binary aBlock. aBlock takes the result of the last evaluation and the next element as arguments. Notable difference from the collection analog is that the streaming variant is a stream of all the intermediate values of the running value."
	"	initialObject	<Object> initial value used as the previous result for the evaluation of the first element
		aBlock	<BlockClosure> binary block combining the value of each element with previous result of its evaluation
		^<CollectingWriteStream>"
	"
		(Array new writing injecting: 0 into: [ :total :each | each + total ]) write: (1 to: 10); conclusion
	"
	| nextObject |
	nextObject := initialObject.
	^self collecting: [:each | nextObject := aBlock cull: nextObject cull: each]! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-private' stamp: 'nice 10/12/2010 00:39'!
isBinary
	"This has to be tricky because Squeak TextConverter is a mess...
	Pretend we are not binary, otherwise TextConverter would attempt writing the Character verbatim..." 
	^false! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-private' stamp: 'nice 10/12/2010 00:44'!
nextPutAll: aString
	self write: aString.
	^aString! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:44'!
rejecting: aBlock
	"Filters written elements using aBlock. aBlock has the same form and semantics as the #reject: block on collections."
	"	aBlock	<BlockClosure>	usual #reject: style block used to filter the elements passing through
		^<TransformWriteStream>"
	"
		(Array new writing rejecting: [ :e | e odd ]) write: (1 to: 10); conclusion
	"
	^self selecting: [:each | (aBlock cull: each) not]! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:44'!
selecting: aBlock
	"Filters written elements using aBlock. aBlock has the same form and semantics as the #select: block on collections."
	"	aBlock	<BlockClosure>	usual #select: style block used to filter the elements passing through
		^<TransformWriteStream>"
	"
		(Array new writing selecting: [ :e | e odd ]) write: (1 to: 10); conclusion
	"
	^self transforming: [:input :output |
		| value |
		[value := input get.
		aBlock cull: value] whileFalse.
		output put: value]! !

!XTWriteStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:44'!
transforming: aBlock
	"This is the most general form of transform stream. The block receives two streams, a virtual stream of written elements (input) and the destination (output). The block can read arbitrary number of elements from input (including none) and write arbitrary number of elements into the output (including none). The block will be invoked as many times as necessary to consume any written elements, or until an Incomplete is raised by the destination.
	Note that if the #contentSpecies of the destination doesn't fit the input of the transformation, the #contentsSpecies of the transform stream has to be set explicitly.
	""	aBlock	<BlockClosure>	binary transformation block that reads elements from input (first argument) and writes elements into output (second argument)
		^<TransformWriteStream>
	""	Convert text into a stream of words
		(Array new writing transforming: [ :in :out || word char |
			word := String new writing.
			[	[  (char := in get) = Character space ] whileFalse: [ word put: char ].
			] ensure: [ out put: (word close; destination) ] ]
		)	write: 'hello world!! bye world!!';
			close;
			terminal
	""	Convert a hex-string into a byte array (2 characters per byte)
		| c2d |
		c2d := [ :char | ('0123456789abcdef' indexOf: char) - 1 ].
		(ByteArray new writing transforming: [ :in :out |
			out put: (c2d value: in get) * 16 + (c2d value: in get) ]
		)	contentsSpecies: String;
			write: '0123456789abcdef';
			close;
			terminal
	"
	^XTTransformWriteStream on: self block: aBlock! !

XTWriteStream subclass: #XTEncodeWriteStream
	instanceVariableNames: 'encoder buffer bufferReading bufferWriting decodedLineEnd encodedLineEnd'
	classVariableNames: ''
	poolDictionaries: 'XTPools'
	category: 'Xtreams-Transforms'!

!XTEncodeWriteStream commentStamp: '<historical>' prior: 0!
Accepts characters, converts them to bytes (0..255) using pre-configured encoding and writes the bytes into destination. Unless set to lineEndTransparent, converts any CRs into configured line-end convention (LF or CRLF).

Instance Variables
	encoder	<Encoder> converts characters to bytes
	buffer	<Buffer on: ByteArray> batch writing buffer
	bufferReading	<ReadStream> read stream on buffer
	bufferWriting	<WriteStream> write stream on buffer
	decodedLineEnd	<Character> CR or nil if transparent
	encodedLineEnd	<ByteArray> encoded bytes of CR in the configured convention (LF or CRLF), or nil if transparent

!

!XTEncodeWriteStream classSide methodsFor: 'instance creation' stamp: 'nice 10/11/2010 21:48'!
on: aDestination encoding: anEncoding
	^self new on: aDestination encoding: anEncoding! !

!XTEncodeWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/31/2010 00:01'!
close
	super close.
	buffer recycle.
	buffer := nil! !

!XTEncodeWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/11/2010 22:07'!
contentsSpecies

	^String! !

!XTEncodeWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/6/2011 22:17'!
encoder

	^encoder! !

!XTEncodeWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/11/2010 22:09'!
on: aSource encoding: anEncoding

	super on: aSource.
	encoder := XTEncoder for: anEncoding.
	buffer := XTRingBuffer new: DefaultBufferSize class: ByteArray.
	bufferReading := buffer reading.
	bufferWriting := buffer writing.
	self setLineEndAuto! !

!XTEncodeWriteStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:08'!
setLineEndAuto

	(Smalltalk os platformName beginsWith: 'Win')
		ifTrue:	[self setLineEndCRLF]
		ifFalse:	[self setLineEndLF]! !

!XTEncodeWriteStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:08'!
setLineEndCRLF

	| stream |
	decodedLineEnd := Character cr.
	stream := ByteArray new writing.
	encoder encode: Character cr on: stream; encode: Character lf on: stream.
	encodedLineEnd := stream contents! !

!XTEncodeWriteStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:08'!
setLineEndLF

	| stream |
	decodedLineEnd := Character cr.
	stream := ByteArray new writing.
	encoder encode: Character lf on: stream.
	encodedLineEnd := stream contents! !

!XTEncodeWriteStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:08'!
setLineEndTransparent

	decodedLineEnd := nil.
	encodedLineEnd := nil! !

!XTEncodeWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 12/31/2010 00:01'!
write: anInteger from: aSequenceableCollection at: startIndex

	| character |
	startIndex to: startIndex + anInteger - 1 do:  [:index |
		buffer writeSize < 10 ifTrue: [destination write: buffer readSize from: bufferReading].
		character := aSequenceableCollection at: index.
		character == decodedLineEnd
			ifTrue:	[bufferWriting write: encodedLineEnd]
			ifFalse:	[encoder encode: character on: bufferWriting]].
	destination write: buffer readSize from: bufferReading.
	^anInteger! !

XTReadStream subclass: #XTCollectReadStream
	instanceVariableNames: 'block cache contentsSpecies direct'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTCollectReadStream commentStamp: '<historical>' prior: 0!
Converts elements being read using the provided conversion block. When the contentSpecies of the source doesn't match the desired contentSpecies of this stream, we optimize bulk reads by reading into an internal buffer.

Instance Variables
	block	<BlockClosure> collecting block (same style as collect: blocks)
	cache	<SequenceableCollection | nil> caches elements before transformation when direct is false
	contentsSpecies	<Class> species for collections of elements of this stream
	direct	<Boolean> are we using an intermediate buffer (when source's contentSpecies doesn't match ours)

!

!XTCollectReadStream classSide methodsFor: 'instance creation' stamp: 'nice 10/10/2010 14:01'!
on: aReadStream block: block
	^self new on: aReadStream block: block! !

!XTCollectReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:50'!
++ anInteger
	^source ++ anInteger! !

!XTCollectReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:50'!
-- anInteger
	^source -- anInteger! !

!XTCollectReadStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 13:58'!
block
	^block! !

!XTCollectReadStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
bufferRead: anInteger into: aSequenceableCollection at: startIndex
	| amount count read |
	count := 0.
	[count < anInteger] whileTrue:
		[amount := (anInteger - count) min: cache size.
		read := [source read: amount into: cache at: 1. amount] on: Incomplete do: [ :ex | ex count ].
		1 to: read do: [:index | aSequenceableCollection at: count + startIndex + index - 1 put: (block value: (cache at: index))].
		count := count + read.
		read < amount ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise]]! !

!XTCollectReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:51'!
close
	super close.
	cache ifNotNil: [ cache recycle ]! !

!XTCollectReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:54'!
contentsSpecies
	^contentsSpecies! !

!XTCollectReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 23:52'!
contentsSpecies: aClass
	contentsSpecies := aClass.
	cache ifNotNil: [ cache recycle ].
	cache := (direct := contentsSpecies == source contentsSpecies)
		ifTrue:	[nil]
		ifFalse:	[
			source contentsSpecies newRecycled: (
				(cache notNil and: [ cache size > 0 ])
					ifTrue: [cache size]
					ifFalse: [DefaultBufferSize])]! !

!XTCollectReadStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
directRead: anInteger into: aSequenceableCollection at: startIndex
	| count |
	count := [source read: anInteger into: aSequenceableCollection at: startIndex. anInteger] on: Incomplete do: [ :ex | ex count ].
	startIndex to: startIndex + count - 1 do: [:index | aSequenceableCollection at: index put: (block value: (aSequenceableCollection at: index))].
	count < anInteger ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise]! !

!XTCollectReadStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 13:58'!
get
	^block value: source get! !

!XTCollectReadStream methodsFor: 'testing' stamp: 'nice 10/10/2010 14:01'!
isPositionable
	^source isPositionable! !

!XTCollectReadStream methodsFor: 'seeking' stamp: 'nice 10/10/2010 14:00'!
length
	^source length! !

!XTCollectReadStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 13:59'!
on: aSource block: aBlock
	self on: aSource.
	block := aBlock.
	contentsSpecies := aSource contentsSpecies.
	direct := true! !

!XTCollectReadStream methodsFor: 'seeking' stamp: 'nice 10/10/2010 14:00'!
position
	^source position! !

!XTCollectReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 23:52'!
position: anInteger
	^source position: anInteger! !

!XTCollectReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 12/30/2010 23:52'!
read: anInteger into: aSequenceableCollection at: startIndex
	direct
		ifTrue:	[self directRead: anInteger into: aSequenceableCollection at: startIndex]
		ifFalse:	[self bufferRead: anInteger into: aSequenceableCollection at: startIndex].
	^anInteger! !

XTReadStream subclass: #XTEncodeReadStream
	instanceVariableNames: 'transparent crPreceeding encoder buffer bufferWriting bufferReading'
	classVariableNames: ''
	poolDictionaries: 'XTPools'
	category: 'Xtreams-Transforms'!

!XTEncodeReadStream commentStamp: '<historical>' prior: 0!
Converts bytes into characters using pre-configured encoding. At the same time, if set to lineEndAuto (default) it can perform line-end translation, converting any line-end convention into CRs. The source stream must provide bytes (0...255).

Instance Variables
	transparent	<Boolean> should the stream perform line-end translations
	crPreceeding	<Boolean> was previous character read a CR (used when not transparent)
	encoder	<Encoder> converts bytes to characters
	buffer	<Buffer on: ByteArray> used to optimize bulk reads
	bufferWriting	<WriteStream> write stream on buffer
	bufferReading	<ReadStream> read stream on buffer

!

!XTEncodeReadStream classSide methodsFor: 'instance creation' stamp: 'nice 10/11/2010 21:47'!
on: aSource encoding: anEncoding
	^self new on: aSource encoding: anEncoding! !

!XTEncodeReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/31/2010 00:00'!
close
	super close.
	buffer recycle.
	buffer := nil! !

!XTEncodeReadStream methodsFor: 'initialize-release' stamp: 'nice 10/11/2010 22:00'!
contentsSpecies
	
	^String! !

!XTEncodeReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/6/2011 22:11'!
encoder

	^encoder! !

!XTEncodeReadStream methodsFor: 'accessing' stamp: 'ul 10/12/2010 01:18'!
get
	| character |
	buffer hasDataToRead ifTrue: [^super get].
	character := encoder decodeFrom: source.
	transparent ifFalse: 
		[character == LF
			ifTrue: [crPreceeding
				ifTrue: 
					[character := encoder decodeFrom: source.
					crPreceeding := character = CR]
				ifFalse: 
					[crPreceeding := false.
					character := CR]]
			ifFalse: [crPreceeding := character = CR]].
	^character! !

!XTEncodeReadStream methodsFor: 'initialize-release' stamp: 'nice 10/11/2010 22:01'!
on: aSource encoding: anEncoding

	super on: aSource.
	encoder := XTEncoder for: anEncoding.
	buffer := XTRingBuffer new: DefaultBufferSize class: ByteArray.
	bufferReading := buffer reading.
	bufferWriting := buffer writing.
	transparent := false.
	crPreceeding := false.! !

!XTEncodeReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
read: anInteger into: aSequenceableCollection at: startIndex

	| remaining position character bufferAvailable |
	remaining := anInteger.
	position := startIndex.
	[remaining > 0] whileTrue: [
		| mark |
		"Top up our buffer if we have room and we need data"
		[bufferWriting write: (buffer writeSize min: remaining) from: source] on: Incomplete do: [:incomplete |
			(incomplete count == 0 and: [buffer hasDataToRead not]) ifTrue: [
				(Incomplete on: aSequenceableCollection count: anInteger - remaining at: startIndex) raise]].

		"We now conduct an inner loop that iterates over the buffer data while:
			a) we need to read more data
			b) there is data available in the buffer
			c) a character can successfully be decoded
		"

		"If our buffer size is too low before we begin our decode loop, we need to take an undo copy in case we cannot decode a character."
		buffer readSize < 10 ifTrue:
			[mark := buffer readPosition.
			encoder backupState ].

		[["The following may raise an incomplete, which means we don't have enough data in the buffer to decode the full character.
		 This is handled by the Incomplete exception capture before."
		character := encoder decodeFrom: bufferReading.

		"If we are not transparent, convert stray LFs in to CRs and CRLFs in to CRs"
		transparent ifFalse: [
			character == LF
				ifTrue:	[character := crPreceeding ifTrue: [nil] ifFalse: [CR]. crPreceeding := false]
				ifFalse:	[crPreceeding := character = CR]].

		"If we didn't filter out an LF at the tail of a CRLF, commit the character to the output."
		character == nil ifFalse:
			[aSequenceableCollection at: position put: character.
			remaining := remaining - 1.
			position := position + 1].

		"Find out how much data we have left in the buffer. If it's too low we need to keep track of the undo record in case we cannot decode a character."
		(bufferAvailable := buffer readSize) < 10 ifTrue:
			[mark := buffer readPosition.
			encoder backupState ].

		remaining > 0 and: [bufferAvailable > 0]] whileTrue]
			on: Incomplete do: [:incomplete |
				"We failed to decode a character, we've hit the end of the buffer and need to refill it. We rewind the buffer and leave the decoding loop
				 to return to the main loop where more data will be fetched in to our buffer."
				buffer readPosition: mark.
				encoder restoreState]].
	^anInteger! !

!XTEncodeReadStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:01'!
setLineEndAuto

	transparent := false! !

!XTEncodeReadStream methodsFor: 'line-end' stamp: 'nice 10/11/2010 22:01'!
setLineEndTransparent

	transparent := true! !

!XTReadStream methodsFor: '*Xtreams-Transforms-private' stamp: 'nice 10/12/2010 00:42'!
basicNext
	"This has to be tricky: a Squeak TextConverter decodes a Character rather than an Integer...
	So let's feed him with a Character."
	^Character value: self get! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:41'!
collecting: aBlock
	"Transform each element using #collect: style block."
	"	aBlock	<BlockClosure>	a #collect: style block used to tranform each element
		^<CollectReadSteam>
	""
		((1 to: 5) reading collecting: [ :e | e * e ]) rest
	""
		((65 to: 90) reading collecting: [ :e | e asCharacter ]) contentsSpecies: String; rest
	"
	^XTCollectReadStream on: self block: aBlock! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:41'!
doing: aBlock
	"Perform and action with each passing element using #do: style block."
	"	aBlock	<BlockClosure>	a #do: style block invoked with each element as it passes through the stream
		^<CollectReadSteam>
	""
		((1 to: 5) reading doing: [ :e | Transcript space; print: e * e ]) rest
	"
	^self collecting: [:each | (aBlock value: each). each]! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/11/2010 22:04'!
encoding: anEncoding
	"Transform bytes into characters using @anEncoding such as #utf8 or #ascii, etc. Any encoding supported by StreamEncoder is allowed.
	The encoding steam also performs automatic line end conversion from arbitrary platform convention to CRs, unless set into a transparent mode"
	"	anEncoding	<Symbol> encoding identifier recognized by StreamEncoder class>>new:
		^<EncodedReadStream>
	""
		((65 to: 90) reading encoding: #ascii) rest
	""
		| crlf text |
		crlf := String with: Character cr with: Character lf.
		text := ('Hello', crlf, 'World') asByteArrayEncoding: #ascii.
		(text reading encoding: #ascii) rest.
		(text reading encoding: #ascii) setLineEndTransparent; rest
	"
	^XTEncodeReadStream on: self encoding: anEncoding! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'MartinKobetic 1/11/2011 17:53'!
encodingBase64
	"Decodes characters of base-64 encoding into bytes. Ignores any intervening whitespace.
	Automatically ends the stream if it encounters final padding characters $=."
	"	^<TransformReadStream>"
	"
		'AAECAwQFBgcICQo= and the rest should be ignored' reading encodingBase64 rest
	"
	| map cache |
	map := [ :char | ('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/' indexOf: char) - 1 ].
	cache := ByteString new: 4.
	^(self transforming: [ :in :out || count end block filter |
		filter := in rejecting: #isSeparator.
		count := [ filter read: 4 into: cache at: 1. 4 ] on: Incomplete do: [ :incomplete | incomplete count].
		count isZero ifTrue: [ Incomplete zero raise ].
		(end := cache indexOf: $=) isZero ifFalse: [ count := count min: end - 1 ].
		count < 2 ifTrue: [ Incomplete zero signal ].
		block := (1 to: 4) inject: 0 into: [ :total :i || sextet |
			sextet := count < i ifTrue: [ 0 ] ifFalse: [ map value: (cache at: i) ].
			sextet negative ifTrue: [ count := i ].
			(total bitShift: 6) + sextet ].
		2 to: count do: [ :i | out put: ((block bitShift: (i - 4) * 8) bitAnd: 255) ].
		count < 4 ifTrue: [ (Incomplete count: count) raise ] ])
			buffer: (XTRingBuffer on: (ByteArray new: 3));
			yourself! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'MartinKobetic 1/17/2011 10:20'!
encodingHex
	"Decodes bytes hex characters."
	"	^<TransformReadStream>"
	"
		(ByteArray withAll: (1 to: 20)) reading encodingHex rest
	"
	| i2c |
	i2c := [ :i | '0123456789abcdef' at: i + 1 ].
	^(self transforming: [ :in :out || byte |
		byte := in get.
		out put: (i2c value: (byte bitShift: -4)).
		out put: (i2c value: (byte bitAnd: 15)) ])
		contentsSpecies: ByteString;
		yourself! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:42'!
injecting: initialObject into: aBlock
	"Accumulates a running value combined with each passing element using the binary aBlock. aBlock takes the result of the last evaluation and the next element as its arguments. Notable difference from the collection analog is that the streaming variant is a stream of all the intermediate values of the running value."
	"	initialObject	<Object> initial value used as the previous result for the evaluation of the first element
		aBlock	<BlockClosure> binary block combining the value of each element with previous result of its evaluation
		^<CollectingReadStream>"
	"
		((1 to: 10) reading injecting: 0 into: [ :total :each | each + total ]) rest
	"
	| nextObject |
	nextObject := initialObject.
	^self collecting: [:each | nextObject := aBlock cull: nextObject cull: each]! !

!XTReadStream methodsFor: '*Xtreams-Transforms-private' stamp: 'nice 10/12/2010 00:43'!
isBinary
	"This has to be tricky because Squeak TextConverter is a mess...
	Pretend we are not binary, otherwise TextConverter would attempt writing the Character verbatim..." 
	^false! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:42'!
rejecting: aBlock
	"Filters elements from the source using aBlock. aBlock has the same form and semantics as the #reject: block on collections."
	"	aBlock	<BlockClosure>	usual #reject: style block used to filter the elements passing through
		^<TransformReadStream>"
	"
		((1 to: 10) reading rejecting: [ :e | e odd ]) rest
	"
	^self transforming: [:input :output |
		| value |
		[value := input get.
		aBlock cull: value] whileTrue.
		output put: value]! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:42'!
selecting: aBlock
	"Filters elements from the source using aBlock. aBlock has the same form and semantics as the #select: block on collections."
	"	aBlock	<BlockClosure>	usual #select: style block used to filter the elements passing through
		^<TransformReadStream>"
	"
		((1 to: 10) reading selecting: [ :e | e odd ]) rest
	"
	^self transforming: [:input :output |
		| value |
		[value := input get.
		aBlock cull: value] whileFalse.
		output put: value]! !

!XTReadStream methodsFor: '*Xtreams-Transforms-transforming' stamp: 'nice 10/10/2010 14:43'!
transforming: aBlock
	"This is the most general form of transform stream. The block receives two streams, the source (input) and a virtual stream of elements to be produced by the stream (output). The block can read arbitrary number of elements from input (including none) and write arbitrary number of elements into the output (including none). The block will be invoked as many times as necessary to produce the required number of elements, or until an Incomplete is raised. Consequently if the block handles Incomplete from the input, it has to raise another Incomplete at some point, otherwise the stream will never end.
	Note that if the contentSpecies of the source doesn't fit the output of the transformation, the contents species of the transform stream has to be set explicitly.
	""	aBlock	<BlockClosure>	binary transformation block that reads elements from input (first argument) and writes elements into output (second argument)
		^<TransformReadStream>
	""	Convert text into a stream of words
		('hello world!! bye world!!' reading transforming: [ :in :out || word char |
			word := String new writing.
			[	[  (char := in get) = Character space ] whileFalse: [ word put: char ].
			] ensure: [ out put: (word close; destination) ] ]
		)	contentsSpecies: Array;
			rest
	""	Convert a hex-string into a byte array (2 characters per byte)
		| c2d |
		c2d := [ :char | ('0123456789abcdef' indexOf: char) - 1 ].
		('0123456789abcdef' reading transforming: [ :in :out |
			out put: (c2d value: in get) * 16 + (c2d value: in get) ]
		)	buffer: (RingBuffer on: (ByteArray new: 1));
			rest
	"
	^XTTransformReadStream on: self block: aBlock! !

XTReadStream subclass: #XTTransformReadStream
	instanceVariableNames: 'buffer bufferWriting sourceAtEnd block closeBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTTransformReadStream commentStamp: 'nice 10/13/2010 20:16' prior: 0!
Transform read stream provides the most expressive form of transformation. The transformation is described by a binary block that is given two arguments, @input and @output. Both @input and @output are streams themselves. The block can read arbitrary amount of elements from @input (including none) and write arbitrary amount of elements into @output (including none).The block will be invoked as many times as necessary to produce the required number of elements, or until an Incomplete is raised. Consequently if the block handles Incomplete from the input, it has to raise another Incomplete at some point, otherwise the stream will never end.

	Note that if the contentSpecies of the source doesn't fit the output of the transformation, the #contentsSpecies of the transform stream has to be set explicitly. The #contentSpecies determines the type of collection employed by the internal buffer that is used as the storage for the virtual output stream of the transformation block. In some cases it might be desirable to tweak the buffering strategy of the virtual output stream. In that case the buffer of the stream can be set manually to any kind of Buffer that matches the requirements of the transformation.
	The closing behavior of the stream can be customized through the closeBlock. The default closeBlock simply propagates #close to the source as with any other transform stream.

Instance Variables
	buffer	<Buffer> holds the contents of the virtual ouput stream
	bufferWriting	<BufferWriteStream> the virtual output stream that is passed to the transformation block as the second argument
	sourceAtEnd	<Boolean> flag indicating that the source has signalled an Incomplete
	block	<BlockClosure> binary transformation block that reads elements from input (first argument) and writes elements into output (second argument)
	closeBlock	<BlockClosure> binary block invoked in response to the #close message, allows customizing the close behavior

!

!XTTransformReadStream classSide methodsFor: 'instance creation' stamp: 'nice 10/10/2010 14:28'!
on: aReadStream block: block
	^self new on: aReadStream block: block! !

!XTTransformReadStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:29'!
block
	^block! !

!XTTransformReadStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:30'!
buffer
	^buffer! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/31/2010 00:02'!
buffer: aBuffer
	buffer recycle.
	buffer := aBuffer.
	bufferWriting := buffer writing! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/31/2010 00:02'!
close
	buffer ifNil: [^self].
	sourceAtEnd := true.
	closeBlock cull: source cull: self.
	buffer recycle.
	buffer := nil! !

!XTTransformReadStream methodsFor: 'accessing' stamp: 'nice 10/10/2010 14:30'!
closeBlock
	^closeBlock! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:32'!
closeBlock: anObject
	closeBlock := anObject! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:32'!
contentsSpecies
	^buffer contentsSpecies! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:33'!
contentsSpecies: aClass

	self buffer: (buffer class new: buffer cacheSize class: aClass)! !

!XTTransformReadStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
fillBuffer: anInteger
"
	We have exhausted our buffer, so fill it up again from our inner stream. Do not read more than is available from the inner stream or more than was originally requested. But try to be as greedy as possible, since filling our buffer is faster than always re-filling our buffer with one object at a time. Also, cap it at @bufferSize, because if we're reading from in-memory or on-disk sources, all the data will be immediately available - which could be gigs worth. We don't want to pull all of that in to our buffer.
	The semantics of our block are interesting. We give it both an input and an output stream. The block must write something to the output stream otherwise we assume this means we're at the end of the stream. It does not need to read from the input stream, but if it does and that stream fires EndOfStream, that also marks us as end of stream.
"
	| count available |
	count := 0.
	[[available := buffer readSize.
	block value: source value: bufferWriting.
	count := count + (buffer readSize - available).
	count >= anInteger ifTrue: [^self].
	buffer hasSpaceToWrite ifFalse: [^self]]
		repeat]
			on: Incomplete do: [:ex | ^sourceAtEnd := true].! !

!XTTransformReadStream methodsFor: 'private' stamp: 'nice 10/10/2010 14:40'!
fillBufferIfRequired: anInteger
"
	If we have any data available in our buffer, return true. If not, attempt to fill up our buffer and return true if that worked. Try not to fill the buffer more than was originally requested (@anInteger), but we make sure we only fill up to what's available from our inner stream. We don't guarantee we'll fill our buffer up to @anInteger's worth, just that we'll attempt to fill the buffer.
"
	buffer ifNil: [^false].
	buffer hasDataToRead ifTrue: [^true].
	sourceAtEnd ifTrue: [^false].
	self fillBuffer: anInteger.
	^buffer hasDataToRead! !

!XTTransformReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
get
	(self fillBufferIfRequired: 1) ifFalse: [Incomplete zero raise].
	^buffer get! !

!XTTransformReadStream methodsFor: 'initialize-release' stamp: 'nice 10/10/2010 14:33'!
on: aReadStream block: aBlock
	
	self on: aReadStream.
	block := aBlock.
	closeBlock := [ :sourceStream | sourceStream close].
	sourceAtEnd := false.
	buffer := XTElasticBuffer new: DefaultBufferSize class: aReadStream contentsSpecies.
	bufferWriting := buffer writing! !

!XTTransformReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
read: anInteger into: aSequenceableCollection at: startIndex
	| count read |
	anInteger isZero ifTrue: [^0].
	count := 0.
	[count < anInteger] whileTrue:
		[(self fillBufferIfRequired: anInteger - count) ifFalse: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise].
		read := (anInteger - count) min: buffer readSize.
		buffer read: read into: aSequenceableCollection at: startIndex + count.
		count := count + read].
	^anInteger! !

!XTTransformReadStream methodsFor: 'printing' stamp: 'nice 10/10/2010 14:33'!
streamingPrintOn: aStream
	super streamingPrintOn: aStream.
	aStream
		space;
		write: '#';
		write: block method homeMethod selector;
		write: ' buffered: ';
		print: buffer readSize.
	buffer readSize isZero ifTrue: [^self].
	aStream
		cr; tab;
		print: buffer contentsFuture! !

XTEncoder subclass: #XTASCIIEncoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTASCIIEncoder commentStamp: 'nice 10/13/2010 20:40' prior: 0!
Implements ASCII encoding.

!

!XTASCIIEncoder classSide methodsFor: 'class initialization' stamp: 'nice 10/13/2010 20:42'!
initialize

	self encoders
		at: #ascii put: self new;
		at: #ASCII put: self new! !

!XTASCIIEncoder methodsFor: 'accessing' stamp: 'nice 10/11/2010 22:02'!
decodeFrom: aReadStream

	| value |
	^(value := aReadStream get) < 128
		ifTrue: [ Character value: value ]
		ifFalse: [ self error: 'Illegal value!!' ]! !

!XTASCIIEncoder methodsFor: 'accessing' stamp: 'nice 10/11/2010 22:02'!
encode: aCharacter on: aWriteStream

	| value |
	(value := aCharacter asInteger) < 128
		ifTrue: [ aWriteStream put: value ]
		ifFalse: [ self error: 'Illegal character!!' ]! !

XTEncoder subclass: #XTISO8859L1Encoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Transforms'!

!XTISO8859L1Encoder commentStamp: '<historical>' prior: 0!
Implements ISO-8859-1 (Latin 1) encoding.

!

!XTISO8859L1Encoder classSide methodsFor: 'class initialization' stamp: 'MartinKobetic 1/6/2011 22:18'!
initialize
	| newSelf |
	newSelf := self new.
	#(#'latin-1' #latin1 #iso8859L1 #ISO8859L1) do: [:id | self encoders at: id put: newSelf]! !

!XTISO8859L1Encoder methodsFor: 'accessing' stamp: 'nice 10/15/2010 21:28'!
decodeFrom: aReadStream
	^Character value: aReadStream get! !

!XTISO8859L1Encoder methodsFor: 'accessing' stamp: 'nice 10/15/2010 21:28'!
encode: aCharacter on: aWriteStream
	aWriteStream put: aCharacter asInteger! !
XTASCIIEncoder initialize!
XTISO8859L1Encoder initialize!
